\ memory.f      - isforth dynamic memory sub-allocator/deallocator
\ ------------------------------------------------------------------------

.( loading memory.f ) cr

\ ------------------------------------------------------------------------

\ any time you allocate memory with a <mmap> syscall you are going to get
\ a block of memory that is a multiple of 4k in size.  this means that if 
\ you repeatedly try to allocate three byes of memory you are going to be 
\ repeatedly given 4096 bytes.  you will be using only 3 bytes in each of
\ these blocks so you will be wasting 4093 bytes of memory with each 
\ allocation.
\
\ this module allocates heaps of memory and on request sub-allocates 
\ pieces parts of those heaps with a granularity of 16 bytes.  if you 
\ require a block of memory that is larger than the largest free block 
\ within any available heap this code will allocate a new heap and give 
\ you a piece (or all) of that instead.
\
\ as you free up allocated blocks the code merges them with adjacent
\ free blocks within the same heap.  this helps prevent fragmentation.
\ when you free up all the allocated blocks within a heap this code will
\ return that heap to the operating system

\ -----------------------------------------------------------------------

 forth definitions

\ ------------------------------------------------------------------------
\ modify these as you see fit (here or at run time)

 $1400000 var heap-size     \ minimum size of heap to allocate (20 megs)

        7 var heap-prot     \ protections to apply to mapped heap (+rwx)
      $22 var heap-flags    \ anonymous, private
    10000 var #blocks       \ max number of allocated blocks per heap

\ 2.2.x kernels do not support anonymous and shared for the same block
\ you must specify private for anonymous mappings.

\ ------------------------------------------------------------------------
\ linked list of heaps allocated from linux via <mmap> syscall

\ these should be headerless but optional - meminfo.f needs to be able
\ to see them

struct: heap
  l-list bfield h.link      \ links
  1 dfield h.map            \ memory map address
  1 dfield h.size           \ memory map size
  1 dfield h.bsize          \ size of mem-blk buffer memory map
  1 dfield h.blocks         \ pointer to buffer for mem-blks
  1 dfield h.unused         \ pointer to list of unused mem-blks
  1 dfield h.free           \ pointer to mem-map of unallocated blocks
  1 dfield h.used           \ pointer to mem-map of allocated blocks 
;struct  

\ ------------------------------------------------------------------------
\ memory block descriptor

struct: mem-blk
  l-list bfield b.link      \ links 
  1 dfield b.addr           \ address of this buffer
  1 dfield b.size           \ byte size of this buffer
  1 dfield b.heap           \ which heap this mem-blk belongs to if any
;struct

\ ------------------------------------------------------------------------

 0 var heap0                \ heap list head pointer
 0 var heaps                \ list of unused heap structures

\ ------------------------------------------------------------------------
\ grab some head space memory for a heap structure

: grab-heap     ( --- a1 )
  thead heap 1+ -
  dup heap erase 
  dup !> thead ;

\ ------------------------------------------------------------------------
\ the size of an array of linked lists of mem-blk structures

<headers

 36 const mem-map           \ array of 9 mem-blk lists of various sizes

\ ------------------------------------------------------------------------

: h-0
  defers default
  off> heap0 
  off> heaps ;

\ ------------------------------------------------------------------------
\ calculate mem-map index for a given block size

: ?index            ( size --- index )
  0 8                       \ test for 8 of the 9 possible block sizes              
  do
    dup                     \ don't eat the size
    16 i << negate and      \ generate a size mask (difficult to explain)
    if                      \ mask fit?
      drop i                \ yes - return i
      undo exit
    then
  -1 +loop
  drop 0 ;                  \ 9th possibility, block is 16 bytes exactly

\ ------------------------------------------------------------------------
\ convert mem-map index to an address within mem-map

: >map-n            ( mem-map index --- a1 )
  4* + ;

\ ------------------------------------------------------------------------
\ index to the correct mem-map item based on mem-blk's size

: ?map              ( mem-map mem-blk --- list )
  b.size @ ?index           \ get mem-map index of this block
  >map-n ;                  \ get block list pointer address

\ ------------------------------------------------------------------------
\ attach a mem-blk structure to the correct mem-map list

: add-mem           ( mem-map mem-blk --- )
  dup 0links                \ mem-blk will be the first in the list
  dup>r ?map                \ point to correct mem-map entry
  dup @ ?dup                \ anything attached to this entry yet ?
  if
    r@ list-add             \ make new entry first in list
  then
  r> swap ! ;               \ point map index at new mem-blk

\ we could make this a better best fit algorithm by linking each mem-blk 
\ structure in a smallest to largest sorted order here

\ ------------------------------------------------------------------------
\ create a mem-blk structure for buffer a1 of size n1 in specified heap

: alloc-blk         ( a1 n1 heap --- mem-blk )
  dup>r h.unused @ ?dup     \ any mem-blks available for recycle?
  if
    dup @next dup           \ yes, get next one in list 
    if                      \ is there a next one?
      dup 0prev             \ yes, erase its prev link
    then
    r@ h.unused !           \ make second item the new first
  else
    r@ h.blocks @           \ get address of mem-blk buffer
    dup @ over incr         \ calculate address to new mem-blk
    mem-blk * + 4+
  then

  dup mem-blk erase         \ erase this mem-blk 

  tuck b.size !             \ store address and size of block in structure
  tuck b.addr !             \ return address of structure
  r> over b.heap ! ;        \ point mem-blk at its parent heap

\ ------------------------------------------------------------------------
\ memory map a block of memory of n1 bytes

: @map              ( n1 --- a1 t | f )
  >r                        \ save size
  0 dup                     \ no offset or fd
  heap-flags                \ protections on memory, normally +rwx
  heap-prot                 \ normally anonymous, shared
  r> 0 <mmap>               \ size, addr
  dup -1 = ;

\ ------------------------------------------------------------------------
\ create a new heap structure 

: ((alloc-heap))    ( --- heap )
  heaps ?dup                \ any heaps to recycle ?
  if
    dup @next dup           \ yes, get next on in list 
    if                      \ is there a next one?
      dup 0prev             \ yes, erase its prev link
    then
    !> heaps                \ make second item the new first
  else
    grab-heap               \ no - allocate a new heap structure
  then

  dup heap erase ;

\ ------------------------------------------------------------------------
\ allocate and initialize a new heap

: (alloc-heap)      ( size --- f1 )
  dup @map                  \ try allocate buffer for heap
  if                        \ failed?
    drop false exit         \ heap not allocated
  then

  #blocks mem-blk *         \ calculate size of buffer for mem-blks

  [ 72                      \ plus two mem map buffers of 36 bytes each
  4+                        \ plus var for mem-blk buffer size
  4+ ] literal +            \ plus var for allocated mem-blk count

  4095 + $fffff000 and      \ make it a multiple of page size

  dup @map
  if                        \ failed?
    drop <munmap> drop      \ free up above
    false exit
  then

  ((alloc-heap)) >r         \ allocate heap structure

       dup r@ h.free !      \ point heap at free mem-map
  36 + dup r@ h.used !      \ point heap at allocated mem-map
  36 + tuck !               \ store size of this buffer 'in' this buffer   
   4 + r@ h.blocks !        \ point heap at mem-blk buffer

  2dup 
  r@ h.map ! r@ h.size !    \ store address and size heap buffer mapping

  swap r@ alloc-blk         \ allocate a mem-blk owning entire heap
  r@ h.free @ swap add-mem  \ add this block to free mem map

  r> heap0 ?dup             \ add new heap to heap list
  if
    over list-add
  then
  !> heap0 true ;           \ success!

\ ------------------------------------------------------------------------
\ allow for allocation of a block larger than the default heap size

: alloc-heap        ( size --- f1 )
  heap-size max
  (alloc-heap) ; 

\ ------------------------------------------------------------------------
\ allocate a heap if we dont already have one (make sure its big enough)

: ?heap             ( size --- size )
  heap0 ?exit               \ allocate a heap if we dont have one
  dup alloc-heap 
  drop ;                    \ assume first heap will allocate ok

\ ------------------------------------------------------------------------
\ split size bytes of memory off mem-blk1 into a new mem-blk

: split-blk         ( mem-blk1 size --- mem-blk2 mem-blk1 )
  >r                        \ move size out of the way
  dup b.addr @              \ get address for new mem-blk
  r@ pluck b.heap @         \ ( addr size heap --- )

  alloc-blk swap            \ allocate a new mem-blk for part we want

  ( mem-blk2 mem-blk1 --- )

  r@ over b.addr +!         \ adjust original block for part we grabbed
  r> negate over b.size +! ;

\ ------------------------------------------------------------------------
\ detach mem-blk from mem-map structure

: grab-blk          ( mem-map mem-blk --- )
  tuck ?map dup>r           \ point to correct map entry for this block
  @ over =                  \ is our block the first one in the list?
  if
    @next dup               \ get next block in list if there is one
    if                      \ is there?
      dup 0prev             \ yes, nullify its prevoius link
    then
    r> !                    \ make item 2 (or null) the first item
  else
    r>drop                  \ dont need mem-map list address now
    list-delete             \ use links within mem-blk to unchain itself
  then ;

\ ------------------------------------------------------------------------
\ scan list of mem-blks within the free mem-map for a 'big enough' block

: scan-free         ( size list --- mem-blk t | f )
  begin
    dup                     \ any (more) memory blocks in this list ?
  while  
    dup b.size @            \ yes, get blocks size
    pluck < not             \ is it big enough?
    if
      nip true exit         \ yes, remember block address
    then
    @next                   \ no, link to next block
  repeat 
  nip ;

\ ------------------------------------------------------------------------
\ search all heaps free map at index for block of requested size or more

\ too much if/and/but loop nesting

: (find-fblk)       ( size index --- size mem-blk t | size f )
  heap0                     \ starting with heap 0
  begin
    dup h.blocks @ @        \ get total number of allocated mem-blks
    #blocks <>              \ are we maxed out on mem-blks?
    if
      dup>r h.free @        \ no - point to free mem-map
      over >map-n @         \ index to correct list in this map
      pluck swap scan-free  \ scan for suitable mem-blk
      if                    \ found one ?
        nip r>drop          \ yes clean up
        true exit           \ return success
      then
      r>                    \ no - get heap back so we can...
    then
    @next dup 0=            \ search next heap at same index - this
  until                     \ makes the algorithm a best-ish fit
  nip ;                     \ see note on add-mem for how to improve this

\ ------------------------------------------------------------------------
\ find a free mem-blk of at least specified size

: find-fblk         ( size index --- size mem-blk t | size f )
  begin
    tuck (find-fblk) not    \ search all heaps at current index
  while                     \ if no block found
    swap 1+                 \ increment the index
    dup 9 =                 \ if index is not yet out of bounds
  until                     \ go back and search all heaps at this index
    drop false              \ all heaps are used up
  else
    rot drop true
  then ;

\ ------------------------------------------------------------------------
\ allocate size bytes of memory from the first bestish block in any heap

headers>

: allocate          ( size --- addr t | f )
  15 + -16 and              \ size must be a multiple of 16 bytes
  ?heap                     \ allocate a heap if we dont have one

  dup ?index                \ get mem-map index for a block this big
  find-fblk not             \ search upwards from this index for a block
  if                        \ that fits requirement
    dup alloc-heap not      \ not found - allocate a new heap
    if                      \ this is the only way an allocation can fail!
      drop false exit       \ panic!
    then
    heap0 h.free @ 32 + @   \ required blk is in new heap - get it
  then 
  
( size mem-blk --- )

  dup b.heap @ dup>r        \ remember which heap this block is from
  h.free @ dup>r            \ remember this heaps free mem-map 
  over grab-blk             \ detach entire block from free mem-map

  2dup b.size @ <>          \ does requested block take up all of 
  if                        \ the located blocks memory?
    swap split-blk          \ no - split off the part we need and
    r> swap add-mem         \ re-attach the bit we dont
  else
    nip r>drop              \ yes - we dont need the free map address now
  then                      \ or the size

  dup b.addr @ dup          \ get address of buffer (ready to return it)
  pluck b.size @ erase      \ erase allocated memory
  swap

  r> h.used @ swap          \ add allocated block to heaps used mem-map
  add-mem true ;

\ ------------------------------------------------------------------------
\ memory deallocation
\ ------------------------------------------------------------------------

 <headers

\ ------------------------------------------------------------------------
\ discard mem-blk - attach it to list of unused mem-blks

: discard-blk       ( mem-blk --- )
  dup 0links                \ block will be the first in the list
  dup b.heap @ h.unused     \ get parent heap of this block
  dup @ ?dup
  if
    pluck list-add          \ yes, link block to head of list
  then
  ! ;                       \ point to new head of list

\ ------------------------------------------------------------------------
\ discard heap - attach it to list of unused heaps

: discard-heap       ( heap --- )
  dup heap0 =               \ is heap being deleted first one in the list?
  if
    heap0 @next dup         \ yes get the next one
    if
      dup 0prev
    then
    !> heap0
  else
    dup list-delete   
  then

  heaps ?dup                \ any blocks already in the list?
  if
    over list-add           \ yes, link block to head of list
  then
  !> heaps ;                \ point to new head of list

\ ------------------------------------------------------------------------
\ return heap buffers to operating system and discard heap structure

: free-heap     ( heap --- )
  dup>r h.used @ 36 + @     \ get size of mem-blk buffer
  r@ h.free @               \ get address of mem-blk buffer  
  <munmap> drop             \ unmap it (it worked :) 
  r@ h.size @ r@ h.map @    \ get heap buffer size and address
  <munmap> drop             \ unmap it
  r> discard-heap ;         \ discard heap structure

\ ------------------------------------------------------------------------
\ merge block with one (adjascent) higher up in memory if possible

: merge-up          ( mem-blk list --- mem-blk t | f )
  over dup b.addr @         \ get end address +1 of our mem-blk
  swap b.size @ + swap
  begin
    dup                     \ any blocks in this list?
  while
    dup b.addr @ pluck =    \ get address of next block, same address as
    if                      \ end of our block?
      nip                   \ yes, discard end address
      dup b.heap @ h.free @ \ get free mem-map of blk to merge with ours
      over grab-blk         \ detach block from free list
      dup b.size @ swap     \ get its size
      discard-blk           \ and discard it
      over b.size +!        \ add its size to our blocks size
      true exit
    then
    @next
  repeat
  nip nip ;

\ ------------------------------------------------------------------------
\ merge block with one (adjascent) lower down in memory if possible

: merge-dn          ( mem-blk list --- mem-blk t | f )
  begin
    dup                     \ any mem-blks in this list?
  while
    dup b.addr @            \ yes, get end address +1 of this block
    over b.size @ +
    pluck b.addr @ =        \ is it the address of our mem-blk?
    if
      over b.size @         \ yes, get size of our block
      rot discard-blk       \ and discard it
      over b.heap @         \ get mem-map this block is attached to
      h.free @
      pluck grab-blk        \ detach merged mem-blk from free list
      over b.size +!        \ extend block into our blocks memory
      true exit
    then
    @next                   \ link to next mem-blk in list
  repeat
  nip ;                     \ discard our mem-blk, leave 0 on stack

\ ------------------------------------------------------------------------

 defer (merge)              \ vectored to one or other of above two 

\ ------------------------------------------------------------------------
\ merge any free memory blocks adjacent to one being freed

: merge             ( mem-blk vector --- mem-blk )
  is (merge)                \ merge-up or merge-dn
  dup b.heap @ h.free @     \ get mem-map of free mem-blks
  9                         \ 9 lists of mem-blks in this mem-map
  for
    2dup @ (merge)          \ search list for mergable mem-blk
    if                      \ found one?
      nip nip               \ yes, clean up
      r>drop exit           \ return merged mem-blk
    then
    4+                      \ point to next list in mem-map
  nxt
  drop ;                    \ no mergable blocks, clean up

\ ------------------------------------------------------------------------
\ scan used mem-map list for block of a given address

: scan-used         ( addr list --- mem-blk t | f )
  dup                       \ are there any items in this mem-map list?
  if
    begin                   \ scan the list then
      2dup b.addr @ =       \ get block address from next mem-blk in list
      if                    \ is it the address of our mem-blk?
        nip true            \ yes, discard address, return success
        exit
      then
      @next dup 0=          \ no,link to next mem-blk in the list
    until                   \ any more items in the list?
  then 
  nip ;                     \ empty list or item not found. return failure

\ ------------------------------------------------------------------------
\ search all lists in used mem-map for block of a given address

: (find-ablk)       ( addr mem-map --- mem-blk t | f )
  9                         \ 9 lists in mem-map to search
  for
    2dup @ scan-used        \ fetch list head from mem-map and scan it
    if                      \ found mem-blk with our address?
      r>drop                \ yes - undo for/nxt
      nip nip               \ clean up (leave mem-blk address only)
      true exit             \ return success
    then
    4+                      \ point to next list head pointer in mem-map
  nxt
  2drop false ;             \ not found, clean up, return failure

\ ------------------------------------------------------------------------
\ search all 'used' mem-maps in all heaps for a block of a given address

: find-ablk         ( addr --- mem-blk t | f )
  heap0                     \ starting with heap 0
  dup 0= ?exit              \ no heaps? return failure

  begin
    2dup h.used @           \ get this heaps mem-map of allocated blocks
    (find-ablk)             \ search it for a block of our address
    if                      \ found?
      nip nip               \ yes, clean up
      true exit             \ return success
    then
    @next dup 0=            \ get next heap from list
  until
  nip ;                     \ block not found, clean up, return failure

\ ------------------------------------------------------------------------

headers>

: free              ( addr --- f1 )
  find-ablk                 \ find allocated mem-blk with this address
  if
    dup b.heap @ dup>r      \ get this blocks heap
    h.used @ over grab-blk  \ detach block from allocated mem-map

    ['] merge-up merge      \ merge block with adjacent free blocks
    ['] merge-dn merge

    dup b.size @            \ get size of block to free
    r@ h.size @ <>          \ same size of heap ?
    if
      r> h.free @           \ no - get this heaps free mem-map
      swap add-mem          \ add block to list of free blocks
    else
      drop r> free-heap     \ yes - discard block, free entire heap
    then

    true exit               \ success
  else 
    false                   \ attempt to free unknown block
  then ;

\ ------------------------------------------------------------------------

 behead

\ ========================================================================
